OBJECT_SCHEMA = {
    'type': 'object',
    'additionalProperties': False,
    'required': ['account_uuid', 'name', 'nodes', 'edges', 'is_inherited'],
    'properties': {
        'account_uuid': {'type': 'string',
                         'format': '[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}'},
        'is_inherited': {'type': 'boolean'},
        'workflow_uuid': {'type': 'string',
                          'format': '[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}',
                          'readOnly': True},
        'trigger_integration_operation_keys': {'type': 'array', 'readOnly': True, 'items': {'type': 'string'}},
        'trigger_operation_uuids': {'type': 'array', 'readOnly': True, 'items': {'type': 'string',
                                                                                 'format': '[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}'}},
        'name': {'type': 'string'},
        'description': {'type': 'string', 'default': None, 'nullable': True},
        'nodes': {
            'type': 'array',
            'items': {
                'type': 'object', 'additionalProperties': False,
                'required': ['node_id', 'name', 'is_trigger', 'integration_uuid', 'operation_key', 'inputs'],
                'properties': {
                    'integration_uuid': {'type': 'string',
                                         'format': '[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}'},
                    'node_uuid': {'type': 'string',
                                  'format': '[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}',
                                  'readOnly': True}, 'node_id': {'type': 'integer'}, 'name': {'type': 'string'},
                    'is_trigger': {'type': 'boolean', 'default': False}, 'integration_key': {'type': 'string'},
                    'operation_key': {'type': 'string'}, 'operation_uuid': {'type': 'string',
                                                                            'format': '[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}',
                                                                            'readOnly': True},
                    'expand_keys': {'default': [], 'type': 'array',
                                    'items': {'type': 'object', 'properties': {'path': {'type': 'string'}}}},
                    'inputs': {'type': 'object'}}}},
        'edges': {
            'type': 'array',
            'items': {
                'type': 'object',
                'additionalProperties': False,
                'required': ['edge_id', 'source_id', 'target_id'],
                'properties': {
                    'edge_uuid': {'type': 'string',
                                  'format': '[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}',
                                  'readOnly': True},
                    'edge_id': {'type': 'integer'}, 'label': {'type': 'string'}, 'source_id': {'type': 'integer'},
                    'source_uuid': {'type': 'string',
                                    'format': '[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}',
                                    'readOnly': True},
                    'target_id': {'type': 'integer'},
                    'target_uuid': {'type': 'string',
                                    'format': '[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}',
                                    'readOnly': True},
                    'expand_keys': {'default': [], 'type': 'array',
                                    'items': {'type': 'object', 'properties': {'path': {'type': 'string'}}}},
                    'filters': {'oneOf': [
                        {'type': 'object', 'additionalProperties': False, 'required': ['operator', 'operations'],
                         'properties': {'operator': {'type': 'string', 'enum': ['and', 'or']},
                                        'operations': {'type': 'array', 'items': {'oneOf': [{'oneOf': [...]},
                                                                                            {'type': 'object',
                                                                                             'additionalProperties': False,
                                                                                             'required': ['left',
                                                                                                          'right',
                                                                                                          'operator'],
                                                                                             'properties': {'left': {
                                                                                                 'oneOf': [
                                                                                                     {'type': 'object',
                                                                                                      'additionalProperties': False,
                                                                                                      'required': [
                                                                                                          'value'],
                                                                                                      'properties': {
                                                                                                          'value': {
                                                                                                              'nullable': True,
                                                                                                              'anyOf': [
                                                                                                                  {
                                                                                                                      'type': 'string'},
                                                                                                                  {
                                                                                                                      'type': 'number'},
                                                                                                                  {
                                                                                                                      'type': 'boolean'},
                                                                                                                  {
                                                                                                                      'type': 'integer'}]},
                                                                                                          'transforms': {
                                                                                                              'type': 'array',
                                                                                                              'items': {
                                                                                                                  'type': 'object',
                                                                                                                  'additionalProperties': False,
                                                                                                                  'required': [
                                                                                                                      'operator',
                                                                                                                      'arguments'],
                                                                                                                  'properties': {
                                                                                                                      'operator': {
                                                                                                                          'type': 'string'},
                                                                                                                      'arguments': {
                                                                                                                          'type': 'object'}}}}}},
                                                                                                     {'type': 'object',
                                                                                                      'additionalProperties': False,
                                                                                                      'required': [
                                                                                                          'path'],
                                                                                                      'properties': {
                                                                                                          'path': {
                                                                                                              'type': 'string'},
                                                                                                          'default': {
                                                                                                              'nullable': True,
                                                                                                              'anyOf': [
                                                                                                                  {
                                                                                                                      'type': 'string'},
                                                                                                                  {
                                                                                                                      'type': 'number'},
                                                                                                                  {
                                                                                                                      'type': 'boolean'},
                                                                                                                  {
                                                                                                                      'type': 'integer'}]},
                                                                                                          'use_first_result': {
                                                                                                              'type': 'boolean',
                                                                                                              'default': True},
                                                                                                          'transforms': {
                                                                                                              'type': 'array',
                                                                                                              'items': {
                                                                                                                  'type': 'object',
                                                                                                                  'additionalProperties': False,
                                                                                                                  'required': [
                                                                                                                      'operator',
                                                                                                                      'arguments'],
                                                                                                                  'properties': {
                                                                                                                      'operator': {
                                                                                                                          'type': 'string'},
                                                                                                                      'arguments': {
                                                                                                                          'type': 'object'}}}}}}]},
                                                                                                            'operator': {
                                                                                                                'type': 'string',
                                                                                                                'enum': [
                                                                                                                    '>',
                                                                                                                    '<',
                                                                                                                    'contains',
                                                                                                                    '>=',
                                                                                                                    '==',
                                                                                                                    '!=',
                                                                                                                    '<=']},
                                                                                                            'right': {
                                                                                                                'oneOf': [
                                                                                                                    {
                                                                                                                        'type': 'object',
                                                                                                                        'additionalProperties': False,
                                                                                                                        'required': [
                                                                                                                            'value'],
                                                                                                                        'properties': {
                                                                                                                            'value': {
                                                                                                                                'nullable': True,
                                                                                                                                'anyOf': [
                                                                                                                                    {
                                                                                                                                        'type': 'string'},
                                                                                                                                    {
                                                                                                                                        'type': 'number'},
                                                                                                                                    {
                                                                                                                                        'type': 'boolean'},
                                                                                                                                    {
                                                                                                                                        'type': 'integer'}]},
                                                                                                                            'transforms': {
                                                                                                                                'type': 'array',
                                                                                                                                'items': {
                                                                                                                                    'type': 'object',
                                                                                                                                    'additionalProperties': False,
                                                                                                                                    'required': [
                                                                                                                                        'operator',
                                                                                                                                        'arguments'],
                                                                                                                                    'properties': {
                                                                                                                                        'operator': {
                                                                                                                                            'type': 'string'},
                                                                                                                                        'arguments': {
                                                                                                                                            'type': 'object'}}}}}},
                                                                                                                    {
                                                                                                                        'type': 'object',
                                                                                                                        'additionalProperties': False,
                                                                                                                        'required': [
                                                                                                                            'path'],
                                                                                                                        'properties': {
                                                                                                                            'path': {
                                                                                                                                'type': 'string'},
                                                                                                                            'default': {
                                                                                                                                'nullable': True,
                                                                                                                                'anyOf': [
                                                                                                                                    {
                                                                                                                                        'type': 'string'},
                                                                                                                                    {
                                                                                                                                        'type': 'number'},
                                                                                                                                    {
                                                                                                                                        'type': 'boolean'},
                                                                                                                                    {
                                                                                                                                        'type': 'integer'}]},
                                                                                                                            'use_first_result': {
                                                                                                                                'type': 'boolean',
                                                                                                                                'default': True},
                                                                                                                            'transforms': {
                                                                                                                                'type': 'array',
                                                                                                                                'items': {
                                                                                                                                    'type': 'object',
                                                                                                                                    'additionalProperties': False,
                                                                                                                                    'required': [
                                                                                                                                        'operator',
                                                                                                                                        'arguments'],
                                                                                                                                    'properties': {
                                                                                                                                        'operator': {
                                                                                                                                            'type': 'string'},
                                                                                                                                        'arguments': {
                                                                                                                                            'type': 'object'}}}}}}]}}}]}}}},
                        {'type': 'object', 'additionalProperties': False, 'required': ['operator', 'operations'],
                         'properties': {'operator': {'type': 'string', 'enum': ['not']},
                                        'operations': {'type': 'array', 'minItems': 1, 'maxItems': 1, 'items': {
                                            'oneOf': [{'oneOf': [...]},
                                                      {'type': 'object', 'additionalProperties': False,
                                                       'required': ['left', 'right', 'operator'], 'properties': {
                                                          'left': {'oneOf': [
                                                              {'type': 'object', 'additionalProperties': False,
                                                               'required': ['value'], 'properties': {
                                                                  'value': {'nullable': True,
                                                                            'anyOf': [{'type': 'string'},
                                                                                      {'type': 'number'},
                                                                                      {'type': 'boolean'},
                                                                                      {'type': 'integer'}]},
                                                                  'transforms': {'type': 'array',
                                                                                 'items': {'type': 'object',
                                                                                           'additionalProperties': False,
                                                                                           'required': ['operator',
                                                                                                        'arguments'],
                                                                                           'properties': {'operator': {
                                                                                               'type': 'string'},
                                                                                                          'arguments': {
                                                                                                              'type': 'object'}}}}}},
                                                              {'type': 'object', 'additionalProperties': False,
                                                               'required': ['path'],
                                                               'properties': {'path': {'type': 'string'},
                                                                              'default': {'nullable': True,
                                                                                          'anyOf': [{'type': 'string'},
                                                                                                    {'type': 'number'},
                                                                                                    {'type': 'boolean'},
                                                                                                    {
                                                                                                        'type': 'integer'}]},
                                                                              'use_first_result': {'type': 'boolean',
                                                                                                   'default': True},
                                                                              'transforms': {'type': 'array',
                                                                                             'items': {'type': 'object',
                                                                                                       'additionalProperties': False,
                                                                                                       'required': [
                                                                                                           'operator',
                                                                                                           'arguments'],
                                                                                                       'properties': {
                                                                                                           'operator': {
                                                                                                               'type': 'string'},
                                                                                                           'arguments': {
                                                                                                               'type': 'object'}}}}}}]},
                                                          'operator': {'type': 'string',
                                                                       'enum': ['>', '<', 'contains', '>=', '==', '!=',
                                                                                '<=']}, 'right': {'oneOf': [
                                                              {'type': 'object', 'additionalProperties': False,
                                                               'required': ['value'], 'properties': {
                                                                  'value': {'nullable': True,
                                                                            'anyOf': [{'type': 'string'},
                                                                                      {'type': 'number'},
                                                                                      {'type': 'boolean'},
                                                                                      {'type': 'integer'}]},
                                                                  'transforms': {'type': 'array',
                                                                                 'items': {'type': 'object',
                                                                                           'additionalProperties': False,
                                                                                           'required': ['operator',
                                                                                                        'arguments'],
                                                                                           'properties': {'operator': {
                                                                                               'type': 'string'},
                                                                                                          'arguments': {
                                                                                                              'type': 'object'}}}}}},
                                                              {'type': 'object', 'additionalProperties': False,
                                                               'required': ['path'],
                                                               'properties': {'path': {'type': 'string'},
                                                                              'default': {'nullable': True,
                                                                                          'anyOf': [{'type': 'string'},
                                                                                                    {'type': 'number'},
                                                                                                    {'type': 'boolean'},
                                                                                                    {
                                                                                                        'type': 'integer'}]},
                                                                              'use_first_result': {'type': 'boolean',
                                                                                                   'default': True},
                                                                              'transforms': {'type': 'array',
                                                                                             'items': {'type': 'object',
                                                                                                       'additionalProperties': False,
                                                                                                       'required': [
                                                                                                           'operator',
                                                                                                           'arguments'],
                                                                                                       'properties': {
                                                                                                           'operator': {
                                                                                                               'type': 'string'},
                                                                                                           'arguments': {
                                                                                                               'type': 'object'}}}}}}]}}}]}}}}]}}}},
        'created_by_user_uuid': {'type': 'string',
                                 'format': '[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}',
                                 'readOnly': True},
        'updated_by_user_uuid': {'type': 'string',
                                 'format': '[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}',
                                 'readOnly': True},
        'created_ts': {'type': 'integer', 'readOnly': True},
        'updated_ts': {'type': 'integer', 'readOnly': True},
        'status_id': {'type': 'integer', 'default': 1},
        'status_name': {'type': 'string', 'readOnly': True},
        'is_active': {'default': True, 'readOnly': True},
        'is_visible': {'default': True, 'readOnly': True}
    }
}



OBJECT_SCHEMA = {
    "type": "object",
    "additionalProperties": False,
    "required": [
        "name",
        "schema",
        "description",
        "account_uuid"
    ],
    "properties": {
        "context_type_uuid": {
            "type": "string",
            "format": "[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}",
            "readOnly": True
        },
        "account_uuid": {
            "type": "string",
            "format": "[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}"
        },
        "content_type": {
            "type": "string",
            "default": "string",
            "enum": [
                "file",
                "string"
            ]
        },
        "name": {
            "type": "string"
        },
        "description": {
            "type": "string"
        },
        "schema": {
            "type": "object"
        },
        "is_inherited": {
            "type": "boolean",
            "default": False
        },
        "is_encrypted": {
            "type": "boolean",
            "default": False
        },
        "version_id": {
            "type": "integer",
            "default": 1,
            "readOnly": True
        },
        "version_uuid": {
            "type": "string",
            "readOnly": True,
            "format": "[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}"
        },
        "created_by_user_uuid": {
            "type": "string",
            "readOnly": True,
            "nullable": True,
            "format": "[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}"
        },
        "updated_by_user_uuid": {
            "type": "string",
            "readOnly": True,
            "nullable": True,
            "format": "[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}"
        },
        "status_id": {
            "type": "integer",
            "default": 1
        },
        "status_name": {
            "type": "string",
            "readOnly": True
        },
        "is_active": {
            "type": "boolean",
            "readOnly": True
        },
        "is_visible": {
            "type": "boolean",
            "readOnly": True
        },
        "created_ts": {
            "type": "integer",
            "readOnly": True
        },
        "updated_ts": {
            "type": "integer",
            "readOnly": True
        }
    }
}



import inspect
from copy import deepcopy
from importlib import reload

from parrot_integrations.core.common import load_integration_module


def validate_integration(integration_key):
    integration = load_integration_module(integration_key=integration_key)
    assert hasattr(integration, 'get_integration_schema')
    assert hasattr(integration, 'connect')
    schema_signature = inspect.signature(integration.get_integration_schema)
    connect_signature = inspect.signature(integration.connect)
    assert len(schema_signature.parameters.keys()) == 0
    assert len(connect_signature.parameters.keys()) == 2
    assert 'extra_attributes' in connect_signature.parameters.keys()
    assert 'credentials' in connect_signature.parameters.keys()
    assert integration.get_integration_schema()


def validate_operation(integration_key, operation_key):
    from parrot_integrations.core.schemas import get_operation_schema
    operation = load_integration_module(integration_key=integration_key, operation_key=operation_key)
    object_schema = None
    if hasattr(operation, 'OBJECT_SCHEMA'):
        object_schema = deepcopy(operation.OBJECT_SCHEMA)
    assert hasattr(operation, 'get_schema')
    assert hasattr(operation, 'process')
    schema_signature = inspect.signature(operation.get_schema)
    process_signature = inspect.signature(operation.process)
    assert len(schema_signature.parameters.keys()) == 0
    schema = operation.get_schema()
    operation_schema = get_operation_schema(integration_key=integration_key, operation_key=operation_key)
    validate_input_schema(inputs=operation_schema['schema']['properties']['inputs'])
    if hasattr(operation, 'OBJECT_SCHEMA'):
        reload(operation)
        assert operation.OBJECT_SCHEMA == object_schema
    assert operation_schema['schema']['type'] == 'object'
    assert operation_schema['schema']['properties']['outputs']
    assert not operation_schema['schema']['additionalProperties']
    kwarg_parameter = any(i.kind.name == 'VAR_KEYWORD' for i in process_signature.parameters.values())
    for keyword in ['workflow_uuid', 'node_uuid', 'processed_ts', 'inputs', 'integration']:
        assert keyword in process_signature.parameters.keys() or kwarg_parameter


def validate_input_schema(inputs):
    print(inputs)
    for i in inputs.get('required', []):
        assert i in inputs['properties'].keys()
    if inputs.get('type') == 'object':
        for k, v in inputs.get('properties', dict()).items():
            print(k, v)
            if v.get('type') == 'object':
                validate_input_schema(inputs=v)
            elif v.get('type') == 'array':
                validate_input_schema(inputs=v['items'])
            elif inputs.get('type'):
                assert isinstance(inputs['type'], str) or isinstance(inputs['type'], list)
            else:
                assert 'oneOf' in v.keys()
    elif inputs.get('type') == 'array':
        validate_input_schema(inputs=inputs['items'])
    elif inputs.get('type'):
        assert isinstance(inputs['type'], str) or isinstance(inputs['type'], list)
    else:
        assert len(inputs['oneOf']) >= 1
        for i in inputs['oneOf']:
            validate_input_schema(inputs=i)



from parrot_integrations.account_service.accounts import OBJECT_SCHEMA
from parrot_integrations.core import search_objects, generate_search_schema


def get_schema():
    search_schema = dict(
        account_type_uuids=dict(
            type='array',
            items=dict(
                type='string',
                format='uuid',
                description='The account type UUIDs to search for'
            )
        ),
        account_uuids=dict(
            type='array',
            items=dict(
                type='string',
                format='uuid',
                description='The account UUIDs to search for'
            )
        ),
        parent_account_uuids=dict(
            type='array',
            items=dict(
                type='string',
                format='uuid',
                description='The account UUIDs to search for'
            )
        ),
        is_active=dict(
            type=['null', 'boolean'],
            default=True
        ),
        is_visible=dict(
            type=['null', 'boolean'],
            default=True
        )
    )
    return generate_search_schema(plural_object_type='accounts', object_schema=OBJECT_SCHEMA,
                                  search_schema=search_schema)


def process(inputs, integration, token, account_uuid, **kwargs):
    account_uuids = inputs.get('account_uuids')
    account_type_uuids = inputs.get('account_type_uuids')
    parent_account_uuids = inputs.get('parent_account_uuids')
    search_parameters = dict(
        account_uuids=account_uuids,
        account_type_uuids=account_type_uuids,
        parent_account_uuids=parent_account_uuids
    )
    if parent_account_uuids is None and account_uuids is None and account_uuid is not None:
        search_parameters['parent_account_uuids'] = [account_uuid]

    return search_objects(
        integration=integration,
        plural_object_type='accounts',
        search_parameters=search_parameters,
        token=token,
        account_uuid=account_uuid
    )



from parrot_integrations.approval_service.comments import OBJECT_SCHEMA
from parrot_integrations.core import get_object, generate_get_schema


def get_schema():
    return generate_get_schema(object_type='comment', object_schema=OBJECT_SCHEMA)


def process(inputs, integration, token, account_uuid, **kwargs):
    return get_object(integration=integration, object_type='comment', object_uuid=inputs['comment_uuid'], token=token,
                      account_uuid=account_uuid)



def evaluate_filter(filters, record):
    operations = []
    for operation in filters.get('operations', []):
        if 'operations' in operation.keys():
            result = evaluate_filter(filters=operation, record=record)
        else:
            result = evaluate_criteria(criteria=operation, record=record)
        operations.append(result)
    return evaluate_operations(operations=operations, operator=filters['operator'])


def evaluate_operations(operations, operator):
    if len(operations) == 0:
        result = True
    elif operator == 'and':
        result = all(operations)
    elif operator == 'or':
        result = any(operations)
    else:
        if len(operations) != 1:
            raise SyntaxError("'not' Operator only works on single operation")
        result = not all(operations)
    return result


def evaluate_criteria(criteria, record):
    from .common import extract_value
    result = False
    try:
        left = extract_value(field=criteria['left'], record=record)
        right = extract_value(field=criteria['right'], record=record)
        if criteria['operator'] == '==':
            result = left == right
        elif criteria['operator'] == '!=':
            result = left != right
        elif criteria['operator'] == '>=':
            result = left >= right
        elif criteria['operator'] == '<=':
            result = left <= right
        elif criteria['operator'] == '<':
            result = left < right
        elif criteria['operator'] == '>':
            result = left > right
        elif criteria['operator'] == 'contains':
            result = right in left
    except Exception:
        pass
    finally:
        return result



def get_schema():
    return dict(
        name='',
        description='',
        is_trigger=False,
        schema=dict(
            type='object',
            additionalProperties=False,
            description='',
            required=['inputs', 'outputs'],
            properties=dict(
                inputs=dict(
                    type='object',
                    additionalProperties=False,
                    required=[],
                    properties=dict(
                    )
                ),
                outputs=dict(
                    type='object',
                    additionalProperties=True,
                    required=[],
                    properties=dict()
                ),
            )
        )
    )


def process(workflow_uuid, node_uuid, processed_ts, inputs, integration, **kwargs):
    pass



from parrot_integrations.core import generate_update_schema, update_object
from parrot_integrations.workflow_service.workflows import OBJECT_SCHEMA


def get_schema():
    return generate_update_schema(object_type='workflow', object_schema=OBJECT_SCHEMA, update_fields=[
        'name',
        'description',
        'is_inherited',
        'status_id',
        'nodes',
        'edges',
    ])


def process(inputs, integration, token, account_uuid, **kwargs):
    return update_object(integration=integration, account_uuid=account_uuid, object_type='workflow',
                         object_uuid=inputs['workflow_uuid'], data=inputs['attributes'], token=token)



OBJECT_SCHEMA = {
    "type": "object",
    "additionalProperties": False,
    "required": [
        "name",
        "integration_type_uuid",
        "account_uuid",
        "extra_attributes"
    ],
    "properties": {
        "account_uuid": {
            "type": "string",
            "format": "[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}"
        },
        "integration_uuid": {
            "type": "string",
            "format": "[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}",
            "readOnly": True
        },
        "integration_type_uuid": {
            "type": "string",
            "format": "[0-9a-fA-F]{8}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{4}\\b-[0-9a-fA-F]{12}"
        },
        "integration_key": {
            "type": "string"
        },
        "name": {
            "type": "string"
        },
        "is_default_integration": {
            "type": "boolean",
            "default": False,
            "writeOnly": True
        },
        "extra_attributes": {
            "type": "object"
        },
        "credentials": {
            "type": "object",
            "writeOnly": True
        },
        "encrypted_credentials": {
            "type": "string"
        },
        "external_attributes": {
            "type": "object",
            "readOnly": True
        },
        "status_id": {
            "type": "integer",
            "default": 1
        },
        "status_name": {
            "type": "string",
            "readOnly": True
        },
        "status_description": {
            "type": "string",
            "readOnly": True
        },
        "is_active": {
            "type": "boolean",
            "readOnly": True
        },
        "is_visible": {
            "type": "boolean",
            "readOnly": True
        },
        "created_ts": {
            "type": "integer",
            "readOnly": True
        },
        "updated_ts": {
            "type": "integer",
            "readOnly": True
        }
    }
}



from croniter import croniter


def get_schema():
    return dict(
        name='Crontab Scheduler',
        description='Trigger Jobs on a Schedule',
        is_trigger=True,
        schema=dict(
            type='object',
            additionalProperties=False,
            description='Trigger Jobs on a Schedule',
            required=['inputs', 'outputs'],
            properties=dict(
                inputs=dict(
                    type='object',
                    additionalProperties=False,
                    properties=dict(
                        expression=dict(
                            type='string',
                        )
                    )
                ),
                outputs=dict(
                    type='object',
                    required=['current_ts'],
                    properties=dict(
                        current_ts=dict(
                            type='number'
                        )
                    )
                ),
            )
        )
    )


def process(inputs, **kwargs):
    output = dict(
        current_ts=inputs['current_ts'].timestamp()
    )
    if croniter.match(inputs['expression'], inputs['current_ts']):
        return output
    else:
        return None



from parrot_integrations.account_service.accounts import OBJECT_SCHEMA
from parrot_integrations.core import generate_trigger_schema, trigger_object


def get_schema():
    return generate_trigger_schema(object_type='account', object_schema=OBJECT_SCHEMA, status='updated')


def process(inputs, **kwargs):
    return trigger_object(inputs=inputs)



from parrot_integrations.core import create_object, generate_create_schema
from parrot_integrations.workflow_service.workflows import OBJECT_SCHEMA


def get_schema():
    return generate_create_schema(object_type='workflow', object_schema=OBJECT_SCHEMA)


def process(inputs, integration, token, account_uuid, **kwargs):
    return create_object(integration=integration, object_type='workflow', data=inputs, token=token,
                         account_uuid=account_uuid)



from parrot_integrations.context_service.context import OBJECT_SCHEMA
from parrot_integrations.core.schemas import format_input_schema


def get_schema():
    return dict(
        name='Upload Context File',
        description='Upload a file to an existing context',
        is_trigger=False,
        schema=dict(
            type='object',
            additionalProperties=False,
            required=['inputs', 'outputs'],
            properties=dict(
                inputs=format_input_schema(
                    input_schema=dict(
                        type='object',
                        additionalProperties=False,
                        required=['context_uuid', 'file'],
                        properties=dict(
                            context_uuid=dict(
                                type='string',
                                format='uuid'
                            ),
                            file=dict(
                                type='string',
                                format='uri'
                            )
                        )
                    )
                ),
                outputs=OBJECT_SCHEMA
            )
        )
    )


def process(integration, inputs, account_uuid, **kwargs):
    return inputs



from parrot_integrations.account_service.account_types import OBJECT_SCHEMA
from parrot_integrations.core import search_objects, generate_search_schema


def get_schema():
    search_schema = dict(
        account_type_uuids=dict(
            type='array',
            items=dict(
                type='string',
                format='uuid',
                description='The account type UUIDs to search for'
            )
        ),
        account_uuids=dict(
            type='array',
            items=dict(
                type='string',
                format='uuid',
                description='The parent account UUIDs to search for'
            )
        ),
        is_active=dict(
            type=['null', 'boolean'],
            default=True
        ),
        is_visible=dict(
            type=['null', 'boolean'],
            default=True
        )
    )
    return generate_search_schema(plural_object_type='account_types', object_schema=OBJECT_SCHEMA,
                                  search_schema=search_schema)


def process(inputs, integration, token, account_uuid, **kwargs):
    return search_objects(
        integration=integration,
        plural_object_type='account_types',
        search_parameters=inputs,
        token=token,
        account_uuid=account_uuid
    )



def get_schema():
    return dict(
        name='Http Request',
        description='Make an HTTP Request',
        is_expandable=False,
        schema=dict(
            type='object',
            additionalProperties=False,
            description='Make an HTTP Request',
            required=['inputs', 'outputs'],
            properties=dict(
                expand_results=dict(
                    type='boolean',
                    default=False,
                    enum=[False]
                ),
                inputs=dict(
                    type='object',
                    additionalProperties=False,
                    required=['method', 'url'],
                    properties=dict(
                        method=dict(
                            type='string',
                        ),
                        url=dict(
                            type='string',
                            format='uri'
                        ),
                        url_params=dict(
                            type='object'
                        ),
                        json_payload=dict(
                            type='object',
                        )
                    )
                ),
                outputs=dict(
                    type='object',
                    required=[
                        'status_code'
                    ],
                    properties=dict(
                        status_code=dict(
                            type='integer',
                        ),
                        json=dict(
                            type='object',
                        ),
                        text=dict(
                            type='string',
                        ),
                    )
                )
            )
        )
    )


def process(inputs, **kwargs):
    import requests

    resp = requests.request(
        method=inputs['method'],
        url=inputs['url'],
        params=dict(
            (param['key'], param['value']) for param in inputs['url_params']
        ) if inputs.get('url_params') else None,
        json=dict(
            (param['key'], param['value']) for param in inputs['json_payload']
        ) if inputs.get('json_payload') else None
    )
    response = dict(
        status_code=resp.status_code,
        text=resp.text
    )
    try:
        response['json'] = resp.json()
    except:
        pass
    return response



import functools
from copy import deepcopy

TRANSFORMS_SCHEMA = dict(
    type="array",
    items={
        "type": "object",
        "additionalProperties": False,
        "required": ["operator", "arguments"],
        "properties": {
            "operator": {
                "type": "string"
            },
            "arguments": {
                "type": "object"
            }
        }
    }
)


@functools.cache
def get_operation_schema(integration_key, operation_key):
    from parrot_integrations.core.common import load_integration_module
    operation = load_integration_module(integration_key, operation_key)
    schema = operation.get_schema()
    schema['schema']['properties']['inputs'] = format_input_schema(
        input_schema=schema['schema']['properties']['inputs'])
    return schema


def generate_get_schema(object_type, object_schema):
    return dict(
        name=f'Get {object_type}',
        description=f'Get a single {object_type} by ID.',
        is_trigger=False,
        schema=dict(
            type='object',
            additionalProperties=False,
            required=['inputs'],
            properties=dict(
                inputs=dict(
                    type='object',
                    additionalProperties=False,
                    required=[f'{object_type}_uuid'],
                    properties={
                        f'{object_type}_uuid': dict(
                            type='string',
                            format='uuid'
                        )
                    }
                ),
                outputs=dict(
                    type='object',
                    additionalProperties=True,
                    required=[object_type, 'exists'],
                    properties={
                        object_type: object_schema,
                        "exists": dict(
                            type='boolean'
                        )
                    }
                ),
            )
        )
    )


def generate_create_schema(object_type, object_schema):
    object_schema = deepcopy(object_schema)
    return dict(
        name=f'Create {object_type}',
        description=f'Create a single {object_type}',
        is_trigger=False,
        schema=dict(
            type='object',
            additionalProperties=False,
            required=['inputs'],
            properties=dict(
                inputs=object_schema,
                outputs=dict(
                    type='object',
                    additionalProperties=True,
                    required=[object_type, 'created'],
                    properties={
                        object_type: object_schema,
                        "created": dict(
                            type='boolean'
                        )
                    }
                ),
            )
        )
    )


def generate_update_schema(object_type, object_schema, update_fields):
    object_schema = deepcopy(object_schema)
    update_schema = dict(
        type='object',
        additionalProperties=False,
        properties={f: object_schema['properties'][f] for f in update_fields}
    )
    return dict(
        name=f'Update {object_type}',
        description=f'Update a single {object_type}',
        is_trigger=False,
        schema=dict(
            type='object',
            additionalProperties=False,
            required=['inputs'],
            properties=dict(
                inputs=dict(
                    type='object',
                    additionalProperties=False,
                    required=[f'{object_type}_uuid', 'attributes'],
                    properties={
                        f'{object_type}_uuid': dict(
                            type='string',
                            format='uuid'
                        ),
                        'attributes': update_schema
                    }
                ),
                outputs=dict(
                    type='object',
                    additionalProperties=True,
                    required=[object_type, 'updated'],
                    properties={
                        object_type: object_schema,
                        "updated": dict(
                            type='boolean'
                        )
                    }
                ),
            )
        )
    )


def generate_search_schema(plural_object_type, object_schema, search_schema):
    return dict(
        name=f'Search {plural_object_type}',
        description=f'Search for {plural_object_type}',
        is_trigger=False,
        schema=dict(
            type='object',
            additionalProperties=False,
            required=['inputs'],
            properties=dict(
                inputs=dict(
                    type='object',
                    additionalProperties=False,
                    properties=search_schema
                ),
                outputs=dict(
                    type='object',
                    additionalProperties=True,
                    required=[plural_object_type],
                    properties={
                        plural_object_type: dict(
                            type='array',
                            items=object_schema,
                        )
                    }
                ),
            )
        )
    )


def generate_trigger_schema(object_type, object_schema, status):
    return dict(
        name=f'{object_type} {status} Trigger',
        description=f'Notify when a {object_type} is {status}',
        is_trigger=True,
        schema=dict(
            type='object',
            additionalProperties=False,
            required=['inputs'],
            properties=dict(
                inputs=dict(
                    type='object',
                    additionalProperties=False,
                    required=['record'],
                    properties=dict(
                        record=
                        dict(
                            oneOf=dict(
                                type='object',
                                required=['path'],
                                additionalProperties=False,
                                properties=dict(
                                    path=dict(
                                        type='string',
                                        enum=['record']
                                    )
                                )
                            )
                        )
                    )
                ),
                outputs=dict(
                    type='object',
                    additionalProperties=True,
                    required=[object_type],
                    properties={
                        object_type: object_schema
                    }
                ),
            )
        )
    )


def format_input_schema(input_schema):
    new_schema = input_schema
    if input_schema['type'] == 'object':
        if not input_schema.get('properties') and input_schema.get('additionalProperties', True):
            new_schema = dict(
                type='array',
                items=format_schema_value(field=None, include_key=True)
            )
        else:
            new_properties = dict()
            for k, v in input_schema['properties'].items():
                if v.get('readOnly'):
                    continue
                if v.get('type') == 'object':
                    new_properties[k] = format_input_schema(input_schema=v)
                elif v.get('type') == 'array':
                    new_properties[k] = v
                    new_properties[k]['items'] = format_input_schema(input_schema=v['items'])
                else:
                    new_properties[k] = format_schema_value(field=v)
            new_schema['properties'] = new_properties
    elif input_schema['type'] == 'array':
        new_schema['items'] = format_input_schema(input_schema=input_schema['items'])
    else:
        new_schema = format_schema_value(field=input_schema)
    return new_schema


def format_schema_value(field=None, include_key=False):
    if field is None:
        field = dict(
            type=['string', 'number', 'integer', 'boolean']
        )
    schema = dict(
        oneOf=[
            dict(
                type='object',
                additionalProperties=False,
                required=['path'],
                properties=dict(
                    path=dict(
                        type='string',
                    ),
                    default=field,
                    transforms=TRANSFORMS_SCHEMA
                )
            ),
            dict(
                type='object',
                additionalProperties=False,
                required=['value'],
                properties=dict(
                    value=field,
                    transforms=TRANSFORMS_SCHEMA
                )
            )
        ]
    )
    if include_key:
        schema['oneOf'][0]['required'].append('key')
        schema['oneOf'][0]['properties']['key'] = dict(type='string')
        schema['oneOf'][1]['required'].append('key')
        schema['oneOf'][1]['properties']['key'] = dict(type='string')
    return schema



from .integration_types import *
from .integrations import *


def get_integration_schema():
    return dict(
        type='object',
        additionalProperties=False,
        description='Integration Service integration',
        required=['extra_attributes', 'credentials'],
        properties=dict(
            extra_attributes=dict(
                type='object',
                additionalProperties=False,
                required=[
                    'base_url'
                ],
                properties=dict(
                    base_url=dict(
                        type='string',
                        description='URL of the Integration Service',
                        default='https://api.example.com/integration/v1/'
                    )
                )
            ),
            credentials=dict(
                type='object',
                additionalProperties=False,
                required=[],
                properties=dict()
            )
        )
    )


def connect(extra_attributes, credentials):
    return dict()



from parrot_integrations.core import generate_trigger_schema, trigger_object
from parrot_integrations.workflow_service.workflows import OBJECT_SCHEMA


def get_schema():
    return generate_trigger_schema(object_type='workflow', object_schema=OBJECT_SCHEMA, status='updated')


def process(inputs, **kwargs):
    return trigger_object(inputs=inputs)



def load_integration_module(integration_key: str, operation_key: str = None):
    from importlib import import_module

    module = import_module(
        name=integration_key if operation_key is None else f'{integration_key}.{operation_key}'
    )
    return module


def format_data(record, schema):
    data = dict()
    for k, v in schema.items():
        if isinstance(v, list):
            value = [format_data(record=record, schema=i) if isinstance(i, dict) else i for i in v]
        elif isinstance(v, dict):
            if all(i in ['value', 'path', 'default', 'transforms'] for i in v.keys()):
                value = extract_value(field=v, record=record)
            else:
                value = {sub_k: extract_value(field=sub_v, record=record) for sub_k, sub_v in v.items()}
        else:
            value = v
        data[k] = value
    return data


def apply_transform(value, transform):
    from parrot_integrations.core import transforms
    value = getattr(transforms, transform['operator'])(value=value, **transform)
    return value


def extract_value(field, record):
    from jsonpath_ng.ext import parse as ng_parse
    value = None
    if 'path' in field.keys():
        result = ng_parse(path=field['path']).find(record)
        if len(result) == 1 or (field.get('use_first_result', True) and len(result) > 1):
            value = result[0].value
        elif len(result) > 1:
            value = [i.value for i in result]
    elif 'value' in field.keys():
        value = field['value']
    else:
        raise SyntaxError("Must use 'path' or 'value'")
    if value is None and field.get('default'):
        value = field['default']
    for transform in field.get('transforms', []):
        value = apply_transform(value=value, transform=transform)
    return value


def list_integrations():
    import pkgutil
    import parrot_integrations
    for loader, name, is_pkg in pkgutil.iter_modules(parrot_integrations.__path__):
        if name not in ['core']:
            yield name


def list_operations(integration_key):
    import pkgutil
    integration_module = load_integration_module(integration_key=integration_key)
    for loader, name, is_pkg in pkgutil.walk_packages(integration_module.__path__, integration_module.__name__ + "."):
        if not is_pkg:
            yield name.replace(integration_key + '.', '')



from urllib.parse import urljoin

import requests

TRANSFORMS_SCHEMA = dict(
    type="array",
    items={
        "type": "object",
        "additionalProperties": False,
        "required": ["operator", "arguments"],
        "properties": {
            "operator": {
                "type": "string"
            },
            "arguments": {
                "type": "object"
            }
        }
    }
)


def get_object(integration, account_uuid, object_type, object_uuid, token, path_override=None, **kwargs):
    from urllib.parse import urljoin
    import requests
    results = {
        object_type: None,
        "exists": False
    }
    path = path_override if path_override is not None else f'/{object_type}s/{object_uuid}'
    resp = requests.get(urljoin(integration['base_url'], path), headers=token)
    if resp.status_code == 200:
        results['exists'] = True
        results[object_type] = resp.json()['response']
    return results


def create_object(integration, account_uuid, object_type, data, token, path_override=None, **kwargs):
    from urllib.parse import urljoin
    import requests
    results = {
        object_type: None,
        "created": False
    }
    path = path_override if path_override is not None else f'/{object_type}s/'
    resp = requests.post(urljoin(integration['base_url'], path), json=data, headers=token)
    if resp.status_code == 200:
        results['created'] = True
        results[object_type] = resp.json()['response']
    return results


def update_object(integration, account_uuid, object_type, object_uuid, data, token, path_override=None, **kwargs):
    from urllib.parse import urljoin
    import requests
    results = {
        object_type: None,
        "updated": False
    }
    path = path_override if path_override is not None else f'/{object_type}s/{object_uuid}'
    resp = requests.patch(urljoin(integration['base_url'], path), json=data, headers=token)
    if resp.status_code == 200:
        results['updated'] = True
        results[object_type] = resp.json()['response']
    return results


def search_objects(integration, search_parameters, plural_object_type, token, limit=100, **kwargs):
    objects = []
    search = True

    page = 1
    while search:
        payload = dict(
            limit=limit,
            search_parameters=search_parameters
        )
        if limit is not None:
            payload['page'] = page
        resp = requests.post(
            url=urljoin(integration['extra_attributes']['base_url'], f'/{plural_object_type}/search'),
            json=payload,
            headers=token
        )
        objects.extend(resp.json()['response'])
        if limit is None or len(resp.json()['response']) < limit:
            search = False
        page += 1
    return {plural_object_type: objects}


def trigger_object(inputs):
    return inputs['record']



from .common import load_integration_module, format_data
from .filters import evaluate_filter
from .graph import validate_workflow_graph
from .integrations import get_object, create_object, update_object, trigger_object, search_objects
from .schemas import generate_create_schema, generate_get_schema, generate_update_schema, generate_trigger_schema, \
    generate_search_schema
